/***************************************************************************
       mt_server.cpp  --  test file for a mulithreaded server

    $Id: mt_server.cpp,v 1.23 2004/08/03 21:01:23 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT

#include <ulxmlrpcpp.h>  // always first header

#include <ulxr_tcpip_connection.h>  // first, don't move: msvc #include bug
#include <ulxr_ssl_connection.h>  
#include <ulxr_http_protocol.h> 

#ifdef __WIN32__
#include <process.h>
#endif

#ifdef __unix__
#include <pthread.h>
#endif

#include <iostream>
#include <memory>

#include <ulxr_except.h>
#include <ulxr_signature.h>
#include <ulxr_mtrpc_server.h>


#ifdef __WIN32__
const unsigned num_threads = 10;
#else
const unsigned num_threads = 100;
#endif

////////////////////////////////////////////////////////////////////////

ulxr::MultiThreadRpcServer *mtServer = 0;

class TestWorker
{
 public:

   TestWorker ()
   {
   }

   ulxr::MethodResponse numThreads (const ulxr::MethodCall &/*calldata*/)
   {
     ulxr::MethodResponse resp;
     resp.setResult(ulxr::Integer(mtServer->numThreads()));
     return resp;
   }

   ulxr::MethodResponse shutdown (const ulxr::MethodCall &/*calldata*/)
   {
     ULXR_COUT << ULXR_PCHAR("TestWorker got signal to shut down\n");
     ulxr::MethodResponse resp;
     resp.setResult(ulxr::Boolean(true));
     ULXR_COUT << ULXR_PCHAR(" Terminating..\n");
     mtServer->terminateAllThreads();
     ULXR_COUT << ULXR_PCHAR(" Returning..\n");
     return resp;
   }

   ulxr::MethodResponse testcall (const ulxr::MethodCall &calldata)
   {
     ulxr::Integer I = calldata.getParam(0);
     int p1 = I.getInteger();
     ULXR_COUT << ULXR_PCHAR("TestWorker got call(")  << p1 << ULXR_PCHAR(") \n");
/*
#ifdef __unix__
     usleep(p1 * 12345 % 1027);
#elif defined(__WIN32__)
     Sleep((p1 * 12345 % 1027) /1000);
#else
#error unsupported platform here
#endif
*/
     ulxr::MethodResponse resp;
     resp.setResult(I);
     return resp;
   }
};


////////////////////////////////////////////////////////////////////////


int main(int argc, char **argv)
{
  try
  {
    ulxr::CppString host = ULXR_PCHAR("localhost");
    if (argc > 1)
      host = ULXR_GET_STRING(argv[1]);

    unsigned port = 32005;
    if (argc > 2)
      port = ulxr_atoi(argv[2]);

    bool wbxml = false;
    if (argc > 3)
      wbxml = ulxr_atoi(argv[3]) != 0;
    
    bool secure = false;
    ulxr::CppString sec = ULXR_PCHAR("unsecured");
    if (argc > 4)
      secure = 0 != ulxr_atoi(argv[4]);
  
    if (secure)  
      sec = ULXR_PCHAR("secured");
  
    ULXR_COUT << ULXR_PCHAR("Serving ") << sec << ULXR_PCHAR(" rpc requests at ") << host << ULXR_PCHAR(":") << port << std::endl;
    ULXR_COUT << ULXR_PCHAR("WBXML: ") << wbxml << std::endl;
    
    std::auto_ptr<ulxr::TcpIpConnection> conn;
#ifdef ULXR_INCLUDE_SSL_STUFF
    if (secure)
    {
      ulxr::SSLConnection *ssl = new ulxr::SSLConnection (true, host, port);
      ssl->setCryptographyData("password", "foo-cert.pem", "foo-cert.pem");
      conn.reset(ssl);
    }
    else
#endif
#ifdef _MSC_VER
	{	 
   	  std::auto_ptr<ulxr::TcpIpConnection> temp(new ulxr::TcpIpConnection (true, host, port));
	  conn = temp;
    }
#else
      conn.reset(new ulxr::TcpIpConnection (true, host, port));
#endif
	  
    std::auto_ptr<ulxr::HttpProtocol> prot(new ulxr::HttpProtocol(conn.get(), conn->getHostName()));
    ulxr::MultiThreadRpcServer server(prot.get(), num_threads, wbxml);
    mtServer = &server;

    TestWorker worker;

    server.addMethod(ulxr::make_method(worker, &TestWorker::testcall),
                     ulxr::Signature(ulxr::Struct()),
                     ULXR_PCHAR("testcall_in_class_dynamic"),
                     ulxr::Signature(ulxr::Integer()),
                     ULXR_PCHAR("Testcase with a dynamic method in a class"));

    server.addMethod(ulxr::make_method(worker, &TestWorker::shutdown),
                     ulxr::Signature(ulxr::Boolean()),
                     ULXR_PCHAR("testcall_shutdown"),
                     ulxr::Signature(),
                     ULXR_PCHAR("Testcase with a dynamic method in a class, shut down server, return old state"));

    server.addMethod(ulxr::make_method(worker, &TestWorker::numThreads),
                     ulxr::Signature(ulxr::Integer()),
                     ULXR_PCHAR("testcall_numthreads"),
                     ulxr::Signature(),
                     ULXR_PCHAR("Returns number of installed threads at startup"));

    unsigned started = server.dispatchAsync();
    ULXR_COUT << ULXR_PCHAR("Started ") << started << ULXR_PCHAR(" threads for dispatching rpc requests\n");
    server.waitAsync(false, true);
    ULXR_COUT << ULXR_PCHAR("Done.\n");
  }

  catch(ulxr::Exception& ex)
  {
    ULXR_COUT << ULXR_PCHAR("Error occured: ") << ULXR_GET_STRING(ex.why()) << std::endl;
    return 1;
  }

  ULXR_COUT << ULXR_PCHAR("Ready.") << std::endl;
  return 0;
}
