/***************************************************************************
          ulxr_connection.cpp  -  provide a connection for rpc-data
                             -------------------
    begin                : Sun Mar 24 2002
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_connection.cpp,v 1.46 2004/06/03 16:57:06 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT
//#define ULXR_SHOW_READ
//#define ULXR_SHOW_WRITE

#include <ulxmlrpcpp.h>  // always first header

#include <ctype.h>
#include <cstdlib>
#include <cerrno>

#ifdef __WIN32__
#include <winsock2.h>
#include <windows.h>
#endif

#ifdef __unix__
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#endif

#include <csignal>
#include <cstdio>

#include "ulxr_connection.h"
#include "ulxr_except.h"


namespace ulxr {


void ULXR_EXPORT Connection::init()
{
  ULXR_TRACE(ULXR_PCHAR("init"));
  fd_handle = -1;
  setTimeout(10);
#if !defined(__WIN32__) && !defined(_WIN32)
  signal (SIGPIPE, SIG_IGN);  // prevent SIGKILL while write()-ing in closing pipe
#endif           
}


ULXR_EXPORT Connection::Connection()
{
  ULXR_TRACE(ULXR_PCHAR("Connection"));
  init();
}


ULXR_EXPORT Connection::~Connection()
{
  ULXR_TRACE(ULXR_PCHAR("~Connection"));
  close();
}


bool ULXR_EXPORT Connection::isOpen() const
{
  ULXR_TRACE(ULXR_PCHAR("isOpen"));
  return fd_handle >= 0;
}


ssize_t ULXR_EXPORT Connection::low_level_write(char const *buff, long len)
{
  ULXR_TRACE(ULXR_PCHAR("low_level_write ") << len);
#ifndef __unix__
      return ::send(fd_handle, buff, len, 0);
#else
      return ::write(fd_handle, buff, len);
#endif

}


void ULXR_EXPORT Connection::write(char const *buff, long len)
{
  ULXR_TRACE(ULXR_PCHAR("write ") << len);
  ULXR_DWRITE_WRITE(buff, len);
  ULXR_DOUT_WRITE(ULXR_PCHAR(""));

  long written;

  if (buff == 0 || !isOpen())
    throw RuntimeException(ApplicationError, i18n(ULXR_PCHAR("Precondition failed for write() call")));

  if (len == 0)
    return;

  fd_set wfd;
  timeval wait;
  wait.tv_sec = timeoutmax;
  wait.tv_usec = 0;

  while (buff != 0 && len > 0) {
    FD_ZERO(&wfd);
    FD_SET((unsigned) fd_handle, &wfd);
    int ready;
    if((ready = select(fd_handle+1, 0, &wfd, 0, &wait)) < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not perform select() call: "))
                                     + getErrorString(getLastError()), 500);

    if(ready == 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Timeout while attempting to write.")), 500);

    if(FD_ISSET(fd_handle, &wfd))
    {
      if ( (written = low_level_write(buff, len)) < 0)
      {
        switch(getLastError())
        {
          case EAGAIN:
          case EINTR:
#ifdef __unix__
            errno = 0;
#endif
          continue;

          case EPIPE:
            close();
            throw ConnectionException(TransportError,
                                       i18n(ULXR_PCHAR("Attempt to write to a connection")
                                       ULXR_PCHAR(" already closed by the peer")), 500);
          /*break; */

          default:
            throw ConnectionException(SystemError,
                                      i18n(ULXR_PCHAR("Could not perform low_level_write() call: ")
                                           + getErrorString(getLastError())), 500);

        }
      }
      else
      {
        buff += written;
        len -= written;
      }
    }
  }
}

    
ssize_t ULXR_EXPORT Connection::low_level_read(char *buff, long len)
{
#ifndef __unix__
    return ::recv(fd_handle, buff, len, 0);
#else
    return ::read(fd_handle, buff, len);
#endif
}


bool ULXR_EXPORT Connection::hasPendingInput() const
{
  return false;
}


ssize_t ULXR_EXPORT Connection::read(char *buff, long len)
{
  long readed = 0;

  ULXR_TRACE(ULXR_PCHAR("read 1"));

  if (buff == 0 || !isOpen())
    throw RuntimeException(ApplicationError,
                           i18n(ULXR_PCHAR("Precondition failed for read() call")));

  ULXR_TRACE(ULXR_PCHAR("read 2 ") << len);

  if (len <= 0)
    return 0;

  ULXR_TRACE(ULXR_PCHAR("read 3"));

  fd_set rfd;
  timeval wait;
  wait.tv_sec = timeoutmax;
  wait.tv_usec = 0;

  FD_ZERO(&rfd);
  FD_SET((unsigned) fd_handle, &rfd);
  int ready;
  if (hasPendingInput())
  {
  }
  else
  {
    if((ready = select(fd_handle+1, &rfd, 0, 0, &wait)) < 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Could not perform select() call: "))
                                + getErrorString(getLastError()), 500);
  
    ULXR_TRACE(ULXR_PCHAR("read 4"));
    if(ready == 0)
      throw ConnectionException(SystemError,
                                i18n(ULXR_PCHAR("Timeout while attempting to read.")), 500);
  }
  
  ULXR_TRACE(ULXR_PCHAR("read 5"));

  if(FD_ISSET(fd_handle, &rfd))
  {
    while ( (readed = low_level_read(buff, len)) < 0)
    {
      switch(getLastError())
      {
        case EAGAIN:
        case EINTR:
#ifdef __unix__
          errno = 0;
#endif
        continue;

        default:
          throw ConnectionException(SystemError,
                                    i18n(ULXR_PCHAR("Could not perform read() call: "))
                                    + getErrorString(getLastError()), 500);
      }
    }
  }

  ULXR_TRACE (ULXR_PCHAR("readed ") << readed);
  ULXR_DWRITE_READ(buff, readed);

  // have Content-length field and got unexpected EOF?
  // otherwise caller gets until EOF
  if (readed == 0 /*&& getBytesToRead() >= 0*/)
  {
    close();
    throw ConnectionException(TransportError,
                               i18n(ULXR_PCHAR("Attempt to read from a connection")
                               ULXR_PCHAR(" already closed by the peer")), 500);
  }

  return readed;
}


void ULXR_EXPORT Connection::cut()
{
  ULXR_TRACE(ULXR_PCHAR("cut"));
  fd_handle = -1;
}


void ULXR_EXPORT Connection::close()
{
  ULXR_TRACE(ULXR_PCHAR("close"));
  if (fd_handle >= 0)
#ifndef __unix__
    ::closesocket(fd_handle);
#else
    ::close(fd_handle);
#endif
  fd_handle = -1;
}


int ULXR_EXPORT Connection::getLastError() {
#ifdef __WIN32__
    return(GetLastError());
#else
    return(errno);
#endif
}


CppString ULXR_EXPORT Connection::getErrorString(int err_number)
{
#ifdef __WIN32__
    LPSTR lpMsgBuf;
    int ok = FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        err_number,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPSTR)&lpMsgBuf,
        0,
        NULL);

    if (ok != 0 && lpMsgBuf != 0)
      return ULXR_GET_STRING(lpMsgBuf);
    else
    {
      char s[40];
      CppString errn  = ULXR_GET_STRING(itoa(err_number, s, 10));
      return i18n(ULXR_PCHAR("Unknown connection problem, Windows error code: #"))+errn;
    }  
#else
    return(ULXR_GET_STRING(strerror(err_number)));
#endif
}


int ULXR_EXPORT Connection::getHandle() const
{
  return fd_handle;
}


};  // namespace ulxr


