/***************************************************************************
                     ulxr_mutex.cpp  -  mutex handling
                             -------------------
    begin                : Thu Dec 03 2002
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_mutex.cpp,v 1.9 2004/01/02 17:47:30 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#include <ulxmlrpcpp.h>  // always first header

#include <cerrno>

#ifdef __unix__
#include <pthread.h>
#endif

#include "ulxr_mutex.h"
#include "ulxr_except.h"


namespace ulxr {


ULXR_EXPORT Mutex::Mutex()
{
#if defined __WIN32__
  if ((handle = CreateMutex(0, false, 0)) < 0)
#else
  if (pthread_mutex_init(&handle, 0) != 0)
#endif
    throw Exception(SystemError, i18n(ULXR_PCHAR("Could not create mutex")));
}


ULXR_EXPORT Mutex::~Mutex() throw()
{
#if defined __WIN32__
  CloseHandle(handle);
#elif _THREAD
  pthread_mutex_destroy(&handle);
#endif
}


void ULXR_EXPORT Mutex::lock()
{
#if defined __WIN32__
  if (WaitForSingleObject(handle, INFINITE) == WAIT_TIMEOUT)
#else
  if (pthread_mutex_lock(&handle) != 0)
#endif
    throw Exception(SystemError, i18n(ULXR_PCHAR("Could not lock mutex")));
}


void ULXR_EXPORT Mutex::unlock()
{
#if defined __WIN32__
  if (!ReleaseMutex(handle))
#else
  if (pthread_mutex_unlock(&handle)!=0)
#endif
    throw Exception(SystemError, i18n(ULXR_PCHAR("Could not unlock mutex")));
}


bool ULXR_EXPORT Mutex::tryLock()
{
#if defined __WIN32__
  if (WaitForSingleObject(handle,0) == WAIT_TIMEOUT)
#else
  int ret = pthread_mutex_trylock(&handle);
  if (ret == EBUSY)
    return false;

  if (ret != 0)
#endif
    throw Exception(SystemError, i18n(ULXR_PCHAR("Could not tryLock mutex")));
  return true;
}


///////////////////////////////////////////////////////////
//


ULXR_EXPORT Mutex::Locker::Locker(Mutex &mtx)
  : mutex(&mtx)
{
  mutex->lock();
}


ULXR_EXPORT Mutex::Locker::~Locker() throw()
{
  try
  {
    mutex->unlock();  // might throw, forget it
  }
  catch(...)
  {
  }
}


};  // namespace ulxr

