
/***************************************************************************
            ulxr_requester.cpp  -  send rpc request ("rpc-client")
                             -------------------
    begin                : Sat Mar 23 2002
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_requester.cpp,v 1.40 2004/08/11 17:26:49 ewald-arnold Exp $

 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

//#define ULXR_DEBUG_OUTPUT
//#define ULXR_SHOW_READ

#include <ulxmlrpcpp.h>  // always first header

#include "ulxr_requester.h"
#include "ulxr_except.h"
#include "ulxr_protocol.h"
#include "ulxr_connection.h"
#include "ulxr_responseparse.h"
#include "ulxr_responseparse_wb.h"

#ifdef __WIN32__
#include <windows.h>
#include <process.h>
#endif

#ifdef __unix__
#include <pthread.h>
#endif

#include <cerrno>
#include <memory>

namespace ulxr {


ULXR_EXPORT Requester::Requester(Protocol* prot, bool wbxml)
{
  wbxml_mode = wbxml;
  protocol = prot;
  cntPendingRequests = 0;
}


void Requester::send_call (const MethodCall& calldata,
                           const CppString &rpc_root)
{
  ULXR_TRACE(ULXR_PCHAR("send_call"));
  if (!protocol->isOpen() )
    protocol->open();
  else
    protocol->resetConnection();

  protocol->setPersistent(false);
  protocol->sendRpcCall(calldata, rpc_root, wbxml_mode);
}


MethodResponse ULXR_EXPORT Requester::waitForResponse()
{
  ULXR_TRACE(ULXR_PCHAR("waitForResponse"));
  return waitForResponse(protocol, wbxml_mode);
}


MethodResponse ULXR_EXPORT
Requester::waitForResponse(Protocol *protocol, bool wbxml_mode)
{
  ULXR_TRACE(ULXR_PCHAR("waitForResponse"));
  char buffer[ULXR_BUFFER_SIZE];
  char *buff_ptr;

  std::auto_ptr<XmlParserBase> parser;
  MethodResponseParserBase *rpb = 0;
  if (wbxml_mode)
  {
    ULXR_TRACE(ULXR_PCHAR("waitForResponse in WBXML"));
    MethodResponseParserWb *rp = new MethodResponseParserWb();
    rpb = rp;
#ifdef _MSC_VER
	std::auto_ptr<XmlParserBase> temp(rp);
	parser = temp;
#else
    parser.reset(rp);
#endif
  }
  else
  {
    ULXR_TRACE(ULXR_PCHAR("waitForResponse in XML"));
    MethodResponseParser *rp = new MethodResponseParser();
    rpb = rp;
#ifdef _MSC_VER
	std::auto_ptr<XmlParserBase> temp(rp);
	parser = temp;
#else
    parser.reset(rp);
#endif
  }

  bool done = false;
  long readed;
  while (!done && protocol->getBytesToRead() != 0
//               && !parser.isComplete()
               && ((readed = protocol->readRaw(buffer, sizeof(buffer))) > 0) )
  {
    buff_ptr = buffer;

    if (protocol->getBytesToRead() == 0
        || parser->isComplete())
      done = true;

    while (readed > 0)
    {
      Protocol::State state = protocol->connectionMachine(buff_ptr, readed);
      if (state == Protocol::ConnError)
      {
        done = true;
        throw ConnectionException(TransportError, i18n(ULXR_PCHAR("network problem occured")), 400);
      }

      else if (state == Protocol::ConnSwitchToBody)
      {
#ifdef ULXR_SHOW_READ
        Cpp8BitString super_data (buff_ptr, readed);
        while ((readed = conn->read(buffer, sizeof(buffer))) > 0)
          super_data.append(buffer, readed);
        ULXR_DOUT_READ(ULXR_PCHAR("superdata 3 start:\n"));

        if (wbxml_mode)
           ULXR_DOUT_READ(binaryDebugOutput(super_data));
        else
          ULXR_DOUT_READ(ULXR_GET_STRING(super_data));

         ULXR_DOUT_READ(ULXR_PCHAR("superdata 3 end:\n") );
#endif
        if (protocol->getBytesToRead() < 0 )
        {
          throw ConnectionException(NotConformingError,
                                    i18n(ULXR_PCHAR("Content-Length of message required")), 411);
        }

        CppString s;
        if (!protocol->responseStatus(s))
          throw ConnectionException(TransportError, s, 500);

      }

      else if (state == Protocol::ConnBody)
      {
        if (!parser->parse(buff_ptr, readed, false))
        {
          throw XmlException(parser->mapToFaultCode(parser->getErrorCode()),
                             i18n(ULXR_PCHAR("Problem while parsing xml response")),
                             parser->getCurrentLineNumber(),
                             ULXR_GET_STRING(parser->getErrorString(parser->getErrorCode())));
        }
        readed = 0;
      }
    }
  }

  if (protocol->isOpen() && !protocol->isPersistent() )
    protocol->close();

  return rpb->getMethodResponse();
}


MethodResponse ULXR_EXPORT
  Requester::call (const MethodCall& calldata, const CppString &rpc_root,
                   const CppString &user, const CppString &pass)
{
   ULXR_TRACE(ULXR_PCHAR("call(..,user, pass)"));
   protocol->setMessageAuthentication(user, pass);
   send_call (calldata, rpc_root);
   return waitForResponse();
}


MethodResponse ULXR_EXPORT
  Requester::call (const MethodCall& calldata, const CppString &rpc_root)
{
   ULXR_TRACE(ULXR_PCHAR("call"));
   send_call (calldata, rpc_root);
   return waitForResponse();
}


/////////////////////////////////////////////////////////////////
//


namespace hidden {


ULXR_EXPORT Receiver_t::Receiver_t(hidden::StaticReceiver_t recv)
 : static_recv(recv)
 , dynamic_recv(0)
{
}


ULXR_EXPORT Receiver_t::Receiver_t(hidden::DynamicReceiver_t recv)
 : static_recv(0)
 , dynamic_recv(recv)
{
}


void ULXR_EXPORT Receiver_t::receive(const MethodResponse &resp)
{
  if (0 != dynamic_recv)
    dynamic_recv->receive(resp);

  else if (0 != static_recv)
    static_recv(resp);
}


void ULXR_EXPORT Receiver_t::free()
{
  if (0 != dynamic_recv)
  {
    delete dynamic_recv;
    dynamic_recv = 0;
  }
}


}  // namespace hidden


/////////////////////////////////////////////////////////////////
//


struct DispatcherData
{
  DispatcherData(Protocol *p, hidden::Receiver_t r, Requester *req)
    : prot(p),
      recv(r),
      requester(req)
    {}

  Requester            *requester;
  Protocol             *prot;
  hidden::Receiver_t    recv;
};


/////////////////////////////////////////////////////////////////
//


void *dispatchThreaded(DispatcherData *data)
{
  data->requester->incPending();
  ULXR_TRACE(ULXR_PCHAR("dispatchThreaded"));
  try
  {
    MethodResponse resp = Requester::waitForResponse(data->prot, data->requester->isWbXml());
    ULXR_TRACE(ULXR_PCHAR("  recv.receive(resp)"));
    data->recv.receive(resp);
    data->requester->decPending();
    data->recv.free();
    delete data->prot;
    delete data;
    return 0;
  }
  catch(...)
  {
    data->requester->decPending();
    delete data->prot;
    delete data;
    return (void*)1;   // FIXME: error handling ??
  }
}


/////////////////////////////////////////////////////////////////
//


void Requester::startDispatch(const MethodCall& call,
                              const CppString &rpc_root,
                              hidden::Receiver_t recv)
{
    ULXR_TRACE(ULXR_PCHAR("startDispatch"));
    send_call(call, rpc_root);
    Protocol *prot = protocol->detach();
    DispatcherData *dd = new DispatcherData (prot, recv, this);

#ifdef __unix__
    typedef void* (*pthread_sig)(void*);
    pthread_t handle;

    bool created = false;
    while (!created)
    {
      created = (0 == pthread_create(&handle, 0, (pthread_sig)dispatchThreaded, dd));
      if (!created)
      {
        switch(protocol->getConnection()->getLastError())
        {
          case EAGAIN: // fallthrough
          case EINTR:
#ifdef __unix__
            errno = 0;
#endif
            continue;
          break;

          default:
          {
            delete prot;
            throw Exception(SystemError,
                            i18n(ULXR_PCHAR("Could not create thread which processes rpc response.\n"
                                            + ULXR_GET_STRING(strerror(prot->getConnection()->getLastError())))));
          }
        }
      }
    }

    unsigned detached = pthread_detach(handle);
    if (detached != 0)
      throw Exception(SystemError,
                      i18n(ULXR_PCHAR("Could not detach thread which processes rpc response")));

#elif defined(__WIN32__)
    typedef unsigned int (__stdcall *thread_sig)(void*);
    unsigned handle;

    unsigned ret = _beginthreadex( 0, 16*1024, (thread_sig)dispatchThreaded,
                                  dd, CREATE_SUSPENDED, &handle );
    int resume = ResumeThread((void*)ret);
    if (resume < 0)
      throw Exception(SystemError,
                      i18n(ULXR_PCHAR("Could not resume thread which processes rpc response")));

    bool created = ret != 0;
    if (!created)
      throw Exception(SystemError,
                      i18n(ULXR_PCHAR("Could not create thread which processes rpc response")));
#else
#error unsupported platform here
#endif
}


void ULXR_EXPORT
Requester::call (const MethodCall& call,
                 const CppString &rpc_root,
                 const CppString &user,
                 const CppString &pass,
                 hidden::Receiver_t recv)
{
   ULXR_TRACE(ULXR_PCHAR("call (.., user, pass, rcv)"));
   protocol->setMessageAuthentication(user, pass);
   startDispatch(call, rpc_root, recv);
}


void ULXR_EXPORT
Requester::call (const MethodCall& call,
                 const CppString &rpc_root,
                 hidden::Receiver_t recv)
{
   ULXR_TRACE(ULXR_PCHAR("call (.., rcv)"));
   startDispatch(call, rpc_root, recv);
}


void ULXR_EXPORT
Requester::transmit (const MethodCall& calldata, const CppString &rpc_root,
                     const CppString &user, const CppString &pass)
{
   ULXR_TRACE(ULXR_PCHAR("transmit (.., user, pass)"));
   protocol->setMessageAuthentication(user, pass);
   protocol->setTransmitOnly();
   send_call(calldata, rpc_root);
}


void ULXR_EXPORT
Requester::transmit (const MethodCall& calldata, const CppString &rpc_root)

{
   ULXR_TRACE(ULXR_PCHAR("transmit"));
   protocol->setTransmitOnly();
   send_call(calldata, rpc_root);
}


int ULXR_EXPORT Requester::numPendingRequests() const
{
  ULXR_TRACE(ULXR_PCHAR("numPendingRequests ") << cntPendingRequests);
  return cntPendingRequests;
}


void ULXR_EXPORT Requester::incPending()
{
  Mutex::Locker lock (pendingMutex);
  ULXR_TRACE(ULXR_PCHAR("incPending ") << cntPendingRequests);
  ++cntPendingRequests;
}


void ULXR_EXPORT Requester::decPending()
{
  Mutex::Locker lock (pendingMutex);
  ULXR_TRACE(ULXR_PCHAR("decPending ") << cntPendingRequests);
  --cntPendingRequests;
}


};  // namespace ulxr
